from itertools import permutations
from typing import Any, Dict

import pytest

from pychoir import (
    LTE,
    All,
    And,
    AreNot,
    Contains,
    ContainsAllOf,
    ContainsAnyOf,
    ContainsNoneOf,
    DictContainsAllOf,
    EqualTo,
    First,
    GreaterThan,
    HasLength,
    InAnyOrder,
    IsEmpty,
    IsEven,
    IsInstance,
    IsNotPresentOr,
    IsOdd,
    Last,
    Len,
    LessThan,
    Matcher,
    NotPresent,
    SetEquals,
    Slice,
)


def test_is_empty():
    assert {'a': []} == {'a': IsEmpty()}
    assert ['foo', '', 'bar'] == ['foo', IsEmpty(), 'bar']
    assert not 'foo' == IsEmpty()

    assert str(IsEmpty()) == 'IsEmpty()'


def test_has_length():
    assert Len is HasLength

    assert {'a': [1, 2, 3]} == {'a': HasLength(3)}
    assert ['123'] == [HasLength(LTE(3))]
    assert not ['1'] == [HasLength(3)]

    assert str(HasLength(1)) == 'HasLength(1)'


def test_all():
    assert {'a': [1, 2, 3]} == {'a': All(IsInstance(int), GreaterThan(0))}
    assert not [-1, 1, 2, 3] == All(GreaterThan(0))

    assert str(All(GreaterThan(0))) == 'All(GreaterThan(0))'


def test_are_not():
    assert [1, 2, 3] == AreNot(IsInstance(str))
    assert not [1, 2, ''] == AreNot(IsInstance(str))

    assert str(AreNot(IsInstance(str))) == 'AreNot(IsInstance(str))'


def test_contains():
    assert {'a': [1, 2, 3]} == {'a': Contains(3)}
    assert not ['12'] == [Contains('3')]

    assert str([Contains('1')]) == "[Contains('1')]"


def test_contains_all_of():
    assert {'a': [1, 2, 3]} == {'a': ContainsAllOf(3, 2, 1)}
    assert not ['12'] == [ContainsAllOf('1', '2', '3')]

    assert str([ContainsAllOf('1', '2', '3')]) == "[ContainsAllOf('1', '2', '3')]"


def test_contains_any_of():
    assert {'a': [1, 2, 3]} == {'a': ContainsAnyOf(5, 3, 1)}
    assert not ['12'] == [ContainsAnyOf('5', '3')]

    assert str([ContainsAnyOf('5', '3')]) == "[ContainsAnyOf('5', '3')]"


def test_contains_none_of():
    assert {'a': [1, 2, 3]} == {'a': ContainsNoneOf(4, 5, 6)}
    assert not ['123'] == [ContainsNoneOf('5', '3', '1')]

    assert str([ContainsNoneOf('5', '3', '1')]) == "[ContainsNoneOf('5', '3', '1')]"


def test_dict_contains_all_of():
    test_input = {'a': [1, 2, 3], 'b': 4, 'c': 'extra'}

    def dict_to_match() -> Dict[Any, Any]:
        return {'a': And(HasLength(3), All(IsInstance(int))), 'b': 4, 'd': NotPresent}

    assert not test_input == dict_to_match()
    assert test_input == DictContainsAllOf(dict_to_match())

    assert not test_input == DictContainsAllOf({'a': [1]})
    assert not test_input == DictContainsAllOf({'a': [1, 2, 3], 'b': 4, 'c': 'extra', 'e': 'not there'})

    assert (str(DictContainsAllOf({'a': And(HasLength(3), All(IsInstance(int))), 'b': 4, 'd': NotPresent}))
            == "DictContainsAllOf({'a': And(HasLength(3), All(IsInstance(int))), 'b': 4, 'd': NotPresent})")


def test_in_any_order():
    assert [1, 2, 3] == InAnyOrder([3, 2, 1])
    assert [1, 2, 2] == InAnyOrder([2, 1, 2])
    for matcher_permutation in permutations([lambda: 3, IsOdd, IsEven]):
        for value_permutation in permutations([1, 2, 3]):
            matcher_iterable = [f() for f in matcher_permutation]  # type: ignore[operator]
            assert value_permutation == InAnyOrder(matcher_iterable)
    assert [1, 2, 2] != InAnyOrder([1, 1, 2])
    assert [1, 2, 3] != InAnyOrder([3, 2])
    assert [1, 2] != InAnyOrder([1, 2, 3])
    assert [1, 2, 3] != InAnyOrder([3, 2, 1, 0])

    assert {1, 2, 3} == InAnyOrder({3, 2, 1})
    assert {1, 2, 3} == InAnyOrder((3, 2, 1))
    assert [{'a': 1}, {'b': 2}] == InAnyOrder([{'b': 2}, {'a': 1}])

    assert str(InAnyOrder([1, 2, 3])) == 'InAnyOrder([1, 2, 3])'


def test_set_equals():
    assert [1, 2, 3] == SetEquals([3, 2, 1])
    assert [1, 2, 3] != SetEquals([3, 2])
    assert [1, 2, 3] != SetEquals([3, 2, 1, 0])
    assert [1, 2, 2] == SetEquals([2, 1, 2])
    assert [1, 2, 2] == SetEquals([1, 1, 2])

    assert {1, 2, 3} == SetEquals({3, 2, 1})
    assert {1, 2, 3} == SetEquals((3, 2, 1))

    assert str(SetEquals([1, 2, 3])) == 'SetEquals([1, 2, 3])'


def test_is_not_present_or():
    with_a = {'a': 1, 'b': 2, 'c': 'whatever'}
    without_a = {'b': 2, 'c': 'whatever'}

    def matcher() -> Matcher:
        return DictContainsAllOf({'a': IsNotPresentOr(1), 'b': 2})

    assert with_a == matcher()
    assert without_a == matcher()

    assert {'a': 2, 'b': 2} != matcher()
    assert {'b': 1} != matcher()

    assert str(matcher()) == "DictContainsAllOf({'a': IsNotPresentOr(1), 'b': 2})"


def test_first():
    assert 'abcdef' == First(3)('abc')
    assert [1, 2, 3, 'a', 'b', 'c'] == First(3)(All(IsInstance(int)))
    assert [1, 2, 3, 'a', 'b', 'c'] != First(4)(All(IsInstance(int)))
    assert [1, 2, 3] == First()(1)

    with pytest.raises(AssertionError):
        assert [1, 2, 3] == First(2)

    with pytest.raises(TypeError) as te_info:
        assert [1, 2, 3] == First(2)([0, 1])([2])  # type: ignore[operator]
    assert str(te_info.value).split('\n')[0] == "'_First' object is not callable"

    with pytest.raises(AssertionError) as exc_info:
        assert [1, 2, 3] == First(2)([0, 1])
    assert str(exc_info.value).split('\n')[0] == 'assert [1, 2, 3] == First(2)([0, 1])[FAILED for [1, 2, 3]]'


def test_last():
    assert 'abcdef' == Last(3)('def')
    assert [1, 2, 3, 'a', 'b', 'c'] == Last(3)(All(IsInstance(str)))
    assert [1, 2, 3, 'a', 'b', 'c'] != Last(4)(All(IsInstance(str)))
    assert [1, 2, 3] == Last()(3)

    with pytest.raises(AssertionError):
        assert [1, 2, 3] == Last(2)

    with pytest.raises(TypeError) as te_info:
        assert [1, 2, 3] == Last(2)([0, 1])([2])  # type: ignore[operator]
    assert str(te_info.value) == "'_Last' object is not callable"

    with pytest.raises(AssertionError) as exc_info:
        assert [1, 2, 3] == Last(2)([1, 2])
    assert str(exc_info.value).split('\n')[0] == 'assert [1, 2, 3] == Last(2)([1, 2])[FAILED for [1, 2, 3]]'


def test_slice():
    assert 'abcdef' == Slice[3:](EqualTo('def'))
    assert 'abcdef' == Slice[:3](All(LessThan('d')))
    assert [1, 2, 3] == Slice[1](EqualTo(2))

    with pytest.raises(AssertionError):
        assert [1, 2, 3] == Slice[2]

    with pytest.raises(TypeError) as te_info:
        assert [1, 2, 3] == Slice[2:3]([0, 1])([2])  # type: ignore[operator]
    assert str(te_info.value) == "'_Slice' object is not callable"

    with pytest.raises(AssertionError) as exc_info:
        assert [1, 2, 3] == Slice[0](EqualTo(0))
    assert (
        str(exc_info.value).split('\n')[0] ==
        'assert [1, 2, 3] == Slice[0](EqualTo(0)[FAILED for 1])[FAILED for [1, 2, 3]]'
    )



import sys
from abc import ABC, abstractmethod
from contextlib import contextmanager
from enum import Enum, auto
from itertools import chain
from typing import Any, Callable, Iterator, List, Optional, Tuple, Type, TypeVar, Union

from pychoir.utils import Default, DefaultType, sequence_or_its_only_member

MatchedType = TypeVar('MatchedType', bound=Any)

if sys.version_info >= (3, 8):
    from typing import Protocol

    class Matchable(Protocol):
        """Type for a value that can be matched against using the :code:`==` operator (has the :code:`__eq__` method).
        In the pychoir API, you can typically pass `Matcher` s and/or normal values where `Matchable` s are expected.
        """
        def __eq__(self, other: MatchedType) -> bool:
            ...  # pragma: no cover
else:
    Matchable = Any


if sys.version_info >= (3, 8):
    from typing import final
else:
    CallableT = TypeVar('CallableT', bound=Callable)

    def final(x: CallableT) -> CallableT:
        return x


T = TypeVar('T')


class _MatcherStatus(str, Enum):
    NOT_RUN = 'NOT RUN'
    PASSED = 'PASSED'
    FAILED = 'FAILED'


class _MatcherDirection(Enum):
    EQ = auto()
    NE = auto()

    @classmethod
    def from_mismatch_expected(cls, mismatch_expected: bool) -> '_MatcherDirection':
        return cls.NE if mismatch_expected else cls.EQ


class _MatcherState:
    def __init__(self) -> None:
        self.__status: _MatcherStatus = _MatcherStatus.NOT_RUN
        self.__direction: Optional[_MatcherDirection] = None
        self.__failed_values: List[Any] = []
        self.__nested_calls: List[Matcher] = []

    def update(self, passed: bool, mismatch_expected: bool, value: Any) -> None:
        direction = _MatcherDirection.from_mismatch_expected(mismatch_expected)
        if self.__direction is None:
            self.__direction = direction
        elif self.__direction != direction:
            # pytest assert rewrite flips comparison on failure, let's not count that
            return

        if passed:
            self.__add_success()
        else:
            self.__add_failure(value)

    def add_nested_call(self, matcher: 'Matcher') -> None:
        self.__nested_calls.append(matcher)

    def reset_failures(self) -> None:
        if self.__status == _MatcherStatus.FAILED:
            self.__status = _MatcherStatus.NOT_RUN
            self.__failed_values = []
        self.__reset_nested_failures()

    def __reset_nested_failures(self) -> None:
        for nested_call in self.__nested_calls:
            nested_call._reset_nested_failures()

    @property
    def status(self) -> _MatcherStatus:
        return self.__status

    @property
    def failed_values(self) -> Tuple[Any, ...]:
        return tuple(self.__failed_values)

    @property
    def was_already_run(self) -> bool:
        return self.status != _MatcherStatus.NOT_RUN

    def __add_failure(self, value: Any) -> None:
        self.__status = _MatcherStatus.FAILED
        self.__failed_values.append(value)

    def __add_success(self) -> None:
        if not self.was_already_run:
            self.__status = _MatcherStatus.PASSED
        if not self.__status == _MatcherStatus.FAILED:
            self.__reset_nested_failures()


class _MatcherContext:
    def __init__(self, mismatch_expected: bool, nested_call: bool) -> None:
        self.__mismatch_expected = mismatch_expected
        self.__nested_call = nested_call

    @property
    def mismatch_expected(self) -> bool:
        return self.__mismatch_expected


class Matcher(ABC):
    """The baseclass for all Matchers.

    :param name:
        Only for Matchers whose class name does not match the call that creates them.
        This applies, for example, to Matchers created by :class:`Transformer` s.

        For example :class:`_First(Matcher)` (that is created by :class:`First(Transformer)`) uses
        this to make its name :code:`'First'` instead of :code:`'_First'` in its textual representation.
    """
    def __init__(self, name: Union[DefaultType, Optional[str]] = Default) -> None:
        super().__init__()
        self.__name = self.__class__.__name__ if name is Default else name
        self.__state = _MatcherState()
        self.__context: Optional[_MatcherContext] = None

    @final
    def as_(self, type_: Type[T]) -> T:
        """Change the static type of the Matcher to make it pass type checking."""
        return self  # type: ignore[return-value]

    @final
    def nested_match(
        self,
        matcher: Union['Matcher', Matchable],
        other: MatchedType,
        expect_mismatch: bool = False
    ) -> bool:
        """For evaluating Matchables (or calling Matchers) from inside a Matcher.

        Takes care of passing all necessary context and updating state when matching.

        :param matcher: The value or Matcher to compare against.
        :param other: The value being compared.
        :param expect_mismatch: Set to True when expecting a mismatch (for example in :class:`Not`).
        """
        if self.__context is not None and self.__context.mismatch_expected:
            expect_mismatch = not expect_mismatch

        if isinstance(matcher, Matcher):
            self.__state.add_nested_call(matcher)
            return matcher.matches(other, _MatcherContext(mismatch_expected=expect_mismatch, nested_call=True))
        else:
            return bool(matcher == other)

    @abstractmethod
    def _matches(self, other: MatchedType) -> bool:
        """Returns True when Matcher matches, False otherwise.

        :param other: The value being compared.

        **To be implemented by all Matchers.**
        """
        ...  # pragma: no cover

    @abstractmethod
    def _description(self) -> str:
        """Returns a textual representation of the Matcher's parameters.

        For example in :code:`"EqualTo('foo')"` the :code:`'foo'` is returned by :code:`_description()`.

        **To be implemented by all Matchers.**
        """
        ...  # pragma: no cover

    @final
    def _reset_nested_failures(
        self,
    ) -> None:
        """For resetting failure state of child matchers in case of passing due to other Matchers.

        For example in :class:`Or`, it is enough that one child Matcher passes.
        The matchers tried up to that point should not report failure.
        After a Matcher reports a success, nested failures get reset automatically.

        It is unlikely that you should ever call this from your tests or custom Matchers yourself.
        """
        self.__state.reset_failures()

    @final
    def matches(self, other: MatchedType, context: _MatcherContext) -> bool:
        with self.__set_context(context):
            passed = self._matches(other)

        reported_passed = passed if not context.mismatch_expected else not passed
        self.__state.update(reported_passed, context.mismatch_expected, other)

        return passed

    @final
    def __eq__(self, other: MatchedType) -> bool:
        return self.matches(other, _MatcherContext(mismatch_expected=False, nested_call=False))

    @final
    def __ne__(self, other: MatchedType) -> bool:
        return not self.matches(other, _MatcherContext(mismatch_expected=True, nested_call=False))

    @final
    def __str__(self) -> str:
        return self.__describe()

    @final
    def __repr__(self) -> str:
        """Textual representation of the Matcher.

        Contains info about failures and failed values
        """
        return self.__describe()

    def __and__(self, other: Matchable) -> 'Matcher':
        """Combines several matchers in a similar fashion as :class:`And`

        Usage:
          >>> from pychoir import IsInstance
          >>> 5 == IsInstance(int) & 5
          True
          >>> 5.0 == IsInstance(int) & 5
          False
        """
        return _AndOperator(self, other)

    def __or__(self, other: Matchable) -> 'Matcher':
        """Combines several matchers in a similar fashion as :class:`Or`

        Usage:
          >>> from pychoir import StartsWith
          >>> 'foo' == StartsWith('foo') | StartsWith('bar')
          True
          >>> 'bar' == StartsWith('foo') | StartsWith('bar')
          True
          >>> 'baz' == StartsWith('foo') | StartsWith('bar')
          False
        """
        return _OrOperator(self, other)

    @final
    def __describe(self) -> str:
        if self.__name:
            return f'{self.__name}({self._description()}){self.__status_string()}'
        else:
            return f'({self._description()}){self.__status_string()}'

    @final
    def __status_string(self) -> str:
        failed_value = sequence_or_its_only_member(self.__state.failed_values)
        return f'[FAILED for {failed_value!r}]' if self.__state.status == _MatcherStatus.FAILED else ''

    @final
    @contextmanager
    def __set_context(self, context: _MatcherContext) -> Iterator[None]:
        self.__context = context
        yield
        self.__context = None


class MatchWrapper:
    def __init__(self, value: MatchedType, matcher: Matcher, did_match: bool):
        self.value = value
        self.matcher = matcher
        self.did_match = did_match

    def __bool__(self) -> bool:
        return self.did_match

    def __str__(self) -> str:
        return f'that({self.value!r}).matches({self.matcher})'

    def __repr__(self) -> str:
        return str(self)


class MatcherWrapper:
    def __init__(self, value: MatchedType):
        self.value = value

    def matches(self, matcher: Matcher) -> MatchWrapper:
        """
        :param matcher: The Matcher to compare `that(value)` with
        :return: a truthy value in case `that(value)` passes the given `matcher`
        """
        context = _MatcherContext(mismatch_expected=False, nested_call=False)
        did_match = matcher.matches(self.value, context)
        return MatchWrapper(self.value, matcher, did_match)


def that(value: MatchedType) -> MatcherWrapper:
    """
    A helper for syntactically sugar coating matches instead of using `==`.

    :param value: The value to pass into the Matcher in `MatcherWrapper.matches()`
    :return: A :class:`MatcherWrapper`

    Usage:
      >>> from pychoir import that, GreaterThan
      >>> assert that(3).matches(GreaterThan(2))
      ...
      >>> assert that(1).matches(GreaterThan(2))
      Traceback (most recent call last):
      ...
      AssertionError
    """
    return MatcherWrapper(value)


class Transformer(ABC):
    @final
    def __call__(self, matcher: Matchable) -> Matcher:
        return self.matches(matcher)

    @abstractmethod
    def matches(self, matcher: Matchable) -> Matcher:
        ...  # pragma: no cover


class _AndOperator(Matcher):
    def __init__(self, *matchers: Matchable):
        super().__init__(name=None)
        self.matchers = matchers

    def _matches(self, other: Any) -> bool:
        return all(self.nested_match(matcher, other) for matcher in self.matchers)

    def _description(self) -> str:
        return ' & '.join(map(repr, self.matchers))

    def __and__(self, other: Matchable) -> Matcher:
        if isinstance(other, _AndOperator):
            return _AndOperator(*chain(self.matchers, other.matchers))
        else:
            return _AndOperator(*self.matchers, other)


class _OrOperator(Matcher):
    def __init__(self, *matchers: Matchable):
        super().__init__(name=None)
        self.matchers = matchers

    def _matches(self, other: Any) -> bool:
        return any(self.nested_match(matcher, other) for matcher in self.matchers)

    def _description(self) -> str:
        return ' | '.join(map(repr, self.matchers))

    def __or__(self, other: Matchable) -> Matcher:
        if isinstance(other, _OrOperator):
            return _OrOperator(*chain(self.matchers, other.matchers))
        else:
            return _OrOperator(*self.matchers, other)



from pychoir import IsEven, IsNegative, IsNonNegative, IsOdd, IsPositive


def test_is_even():
    assert 0 == IsEven()
    assert -2 == IsEven()
    assert 2 == IsEven()
    assert 1 != IsEven()

    assert str(IsEven()) == 'IsEven()'


def test_is_odd():
    assert 1 == IsOdd()
    assert -1 == IsOdd()
    assert 0 != IsOdd()
    assert -2 != IsOdd()

    assert str(IsOdd()) == 'IsOdd()'


def test_is_positive():
    assert 1 == IsPositive()
    assert 2 == IsPositive()
    assert 0 != IsPositive()
    assert -1 != IsPositive()

    assert str(IsPositive()) == 'IsPositive()'


def test_is_non_negative():
    assert 1 == IsNonNegative()
    assert 0 == IsNonNegative()
    assert -1 != IsNonNegative()

    assert str(IsNonNegative()) == 'IsNonNegative()'


def test_is_negative():
    assert -1 == IsNegative()
    assert 0 != IsNegative()
    assert 1 != IsNegative()

    assert str(IsNegative()) == 'IsNegative()'



# type: ignore
from typing import Optional

_DUMMY_VERSION = '0.0.0.dev0'


def version_from_pyproject() -> Optional[str]:
    try:
        import tomli
        from pathlib import Path
        path = Path(__file__).parent.parent / "pyproject.toml"
        with path.open("rb") as f:
            pyproject = tomli.load(f)
        return pyproject["project"]["version"]
    except ImportError:
        return None


version = version_from_pyproject() or _DUMMY_VERSION



from typing import Any

from pychoir.core import Matcher


class EqualTo(Matcher):
    """A somewhat redundant matcher checking for equality. Most Matchers can take
    values and/or Matchers and **you should prefer bare values to wrapping them
    in EqualTo**.

    A notable exception is when used with & and | operators with a bare value on the left-most position

    :param value: The value to compare against.

    Usage:
      >>> from pychoir import All, EqualTo, IsInstance
      >>> [1, 1, 1] == All(EqualTo(1))  # Bad, do not do this
      True
      >>> [1, 1, 1] == All(1)  # Good, do this instead
      True
      >>> 1 == EqualTo(1) & IsInstance(int)  # Needed here, but see below
      True
      >>> 1 == IsInstance(int) & 1  # Bare value works here
      True
    """
    def __init__(self, value: Any):
        super().__init__()
        self.value = value

    def _matches(self, other: Any) -> bool:
        return bool(other == self.value)

    def _description(self) -> str:
        return repr(self.value)


EQ = EqualTo


class NotEqualTo(Matcher):
    """A Matcher checking that compared value is not equal to the given value.

    :param value: The value to compare against.

    Usage:
      >>> from pychoir import NotEqualTo
      >>> 1 == NotEqualTo(2)
      True
      >>> 1 == NotEqualTo(1)
      False
    """
    def __init__(self, value: Any):
        super().__init__()
        self.value = value

    def _matches(self, other: Any) -> bool:
        return bool(other != self.value)

    def _description(self) -> str:
        return repr(self.value)


NE = NotEqualTo


class GreaterThan(Matcher):
    """A Matcher checking that compared value is greater than the given value.

    :param threshold: The value to compare against.

    Usage:
      >>> from pychoir import GreaterThan
      >>> [2] == [GreaterThan(2)]
      False
      >>> [3] == [GreaterThan(2)]
      True
    """
    def __init__(self, threshold: Any):
        super().__init__()
        self.threshold = threshold

    def _matches(self, other: Any) -> bool:
        return bool(other > self.threshold)

    def _description(self) -> str:
        return repr(self.threshold)


GT = GreaterThan


class GreaterThanOrEqualTo(Matcher):
    """A Matcher checking that compared value is greater than or equal to the given value.

    :param threshold: The value to compare against.

    Usage:
      >>> from pychoir import GreaterThanOrEqualTo
      >>> 2 == GreaterThanOrEqualTo(2)
      True
      >>> 1 == GreaterThanOrEqualTo(2)
      False
    """
    def __init__(self, threshold: Any):
        super().__init__()
        self.threshold = threshold

    def _matches(self, other: Any) -> bool:
        return bool(other >= self.threshold)

    def _description(self) -> str:
        return repr(self.threshold)


GTE = GreaterThanOrEqualTo


class LessThan(Matcher):
    """A Matcher checking that compared value is less than the given value.

    :param threshold: The value to compare against.

    Usage:
      >>> from pychoir import LessThan
      >>> 2 == LessThan(2)
      False
      >>> 1 == LessThan(2)
      True
    """
    def __init__(self, threshold: Any):
        super().__init__()
        self.threshold = threshold

    def _matches(self, other: Any) -> bool:
        return bool(other < self.threshold)

    def _description(self) -> str:
        return repr(self.threshold)


LT = LessThan


class LessThanOrEqualTo(Matcher):
    """A Matcher checking that compared value is less than or equal to the given value.

    :param threshold: The value to compare against.

    Usage:
      >>> from pychoir import LessThanOrEqualTo
      >>> 2 == LessThanOrEqualTo(2)
      True
      >>> 3 == LessThanOrEqualTo(2)
      False
    """
    def __init__(self, threshold: Any):
        super().__init__()
        self.threshold = threshold

    def _matches(self, other: Any) -> bool:
        return bool(other <= self.threshold)

    def _description(self) -> str:
        return repr(self.threshold)


LTE = LessThanOrEqualTo



from .core import Matchable, Matcher, that  # noqa: F401  # isort:skip

from .callables import WhenPassedTo  # noqa: F401
from .comparisons import (  # noqa: F401
    EQ,
    GT,
    GTE,
    LT,
    LTE,
    NE,
    EqualTo,
    GreaterThan,
    GreaterThanOrEqualTo,
    LessThan,
    LessThanOrEqualTo,
    NotEqualTo,
)
from .containers import (  # noqa: F401
    All,
    AreNot,
    Contains,
    ContainsAllOf,
    ContainsAnyOf,
    ContainsNoneOf,
    DictContainsAllOf,
    First,
    HasLength,
    InAnyOrder,
    IsEmpty,
    IsNotPresentOr,
    Last,
    Len,
    NotPresent,
    SetEquals,
    Slice,
)
from .existential import (  # noqa: F401
    Anything,
    In,
    Is,
    IsFalsy,
    IsNoneOr,
    IsTruthy,
    OneOf,
    Optionally,
)
from .integration import M, MatcherLike, Matches  # noqa: F401
from .logical import AllOf, And, AnyOf, IsNoneOf, Not, Or, ResultsTrueFor  # noqa: F401
from .numeric import IsEven, IsNegative, IsNonNegative, IsOdd, IsPositive  # noqa: F401
from .strings import EndsWith, MatchesRegex, StartsWith  # noqa: F401
from .types import ConvertsTo, IsInstance, OfType  # noqa: F401



from typing import Any, Callable

from pychoir.core import Matchable, Matcher


class And(Matcher):
    """A Matcher checking that the compared value matches *all* the passed :class:`Matchable` s.

    Consider using the :code:`&` operator (:func:`Matcher.__and__`) instead: :code:`StartsWith('a') & HasLength(3)`.

    :param matchers: The value(s) and/or matcher(s) to compare against.

    Usage:
      >>> from pychoir import And, IsInstance, HasLength, StartsWith
      >>> 5 == And(IsInstance(int), 5)
      True
      >>> 'abc' == And(StartsWith('a'), HasLength(3))
      True
      >>> 4 == And(IsInstance(int), 5)
      False
    """
    def __init__(self, *matchers: Matchable):
        super().__init__()
        self.matchers = matchers

    def _matches(self, other: Any) -> bool:
        return all(self.nested_match(matcher, other) for matcher in self.matchers)

    def _description(self) -> str:
        return ', '.join(map(repr, self.matchers))


AllOf = And


class Or(Matcher):
    """A Matcher checking that the compared value matches *at least one of* the passed :class:`Matchable` s.

    Consider using the :code:`|` operator (:func:`Matcher.__or__`) instead: :code:`StartsWith('a') | StartsWith('b')`.

    :param matchers: The value(s) and/or matcher(s) to compare against.

    Usage:
      >>> from pychoir import Or, IsInstance, StartsWith
      >>> 5 == Or(IsInstance(int), 5)
      True
      >>> 'abc' == Or(StartsWith('abc'), StartsWith('def'))
      True
      >>> '4' == Or(IsInstance(int), 5)
      False
    """
    def __init__(self, *matchers: Matchable):
        super().__init__()
        self.matchers = matchers

    def _matches(self, other: Any) -> bool:
        return any(self.nested_match(matcher, other) for matcher in self.matchers)

    def _description(self) -> str:
        return ', '.join(map(repr, self.matchers))


AnyOf = Or


class Not(Matcher):
    """A Matcher checking that the compared value matches *none of* the passed :class:`Matchable` s.

    :param matchers: The value(s) and/or matcher(s) to compare against.

    Usage:
      >>> from pychoir import Not, IsInstance, StartsWith
      >>> 5 == Not(IsInstance(str))
      True
      >>> 'abc' == Not(StartsWith('abc'), StartsWith('def'))
      False
      >>> 4 == Not(IsInstance(str), 5)
      True
    """
    def __init__(self, *matchers: Matchable):
        super().__init__()
        self.matchers = matchers

    def _matches(self, other: Any) -> bool:
        return not any(self.nested_match(matcher, other, expect_mismatch=True) for matcher in self.matchers)

    def _description(self) -> str:
        return ', '.join(map(repr, self.matchers))


IsNoneOf = Not


class ResultsTrueFor(Matcher):
    """A Matcher checking that the compared value results :code:`True` when given to the passed function(s).

    :param conditions: The functions to check against.

    Usage:
      >>> from pychoir import ResultsTrueFor
      >>> 5 == ResultsTrueFor(bool)
      True
      >>> 'abc' == ResultsTrueFor(lambda x: x[2] == 'c')
      True
    """
    def __init__(self, *conditions: Callable[[Any], bool]):
        super().__init__()
        self.conditions = conditions

    def _matches(self, other: Any) -> bool:
        return all(condition(other) for condition in self.conditions)

    def _description(self) -> str:
        return ', '.join(map(repr, self.conditions))



from typing import Any, List

from pychoir.utils import i_removed, sequence_or_its_only_member


def test_sequence_or_its_first_member():
    empty_list: List[Any] = []
    assert sequence_or_its_only_member(empty_list) is empty_list
    assert sequence_or_its_only_member(['only']) == 'only'
    assert sequence_or_its_only_member(('several', 'members')) == ('several', 'members')


def test_i_removed():
    t = (0, 1, 2, 3, 4, 5)
    assert i_removed(t, 0) == (1, 2, 3, 4, 5)
    assert i_removed(t, 1) == (0, 2, 3, 4, 5)
    assert i_removed(t, 5) == (0, 1, 2, 3, 4)
    assert i_removed(t, 6) == t



import re
import sys
from typing import Any, Union

from pychoir import Matcher

if sys.version_info >= (3, 7):
    from re import Pattern
else:
    Pattern = Any


class StartsWith(Matcher):
    """A Matcher checking that the compared string :code:`.startswith()` the passed string.

    :param start: The string the compared value is expected to start with.

    Usage:
      >>> from pychoir import StartsWith
      >>> 'foobar' == StartsWith('foo')
      True
      >>> 'barbar' == StartsWith('foo')
      False
    """
    def __init__(self, start: str):
        super().__init__()
        self.start = start

    def _matches(self, other: str) -> bool:
        return other.startswith(self.start)

    def _description(self) -> str:
        return repr(self.start)


class EndsWith(Matcher):
    """A Matcher checking that the compared string :code:`.endswith()` the passed string.

    :param end: The string the compared value is expected to end with.

    Usage:
      >>> from pychoir import EndsWith
      >>> 'foobar' == EndsWith('bar')
      True
      >>> 'foofoo' == EndsWith('bar')
      False
    """
    def __init__(self, end: str):
        super().__init__()
        self.end = end

    def _matches(self, other: str) -> bool:
        return other.endswith(self.end)

    def _description(self) -> str:
        return repr(self.end)


class MatchesRegex(Matcher):
    """A Matcher checking that the compared string matches the passed regular expression.

    :param regex: The regular expression (as a string or a :class:`re.Pattern`).

    Usage:
      >>> import re
      >>> from pychoir import MatchesRegex
      >>> 'foobar' == MatchesRegex(r'^f.obar')
      True
      >>> 'foofoo' == MatchesRegex(re.compile(r'^b[ao]r$'))
      False
    """
    def __init__(self, regex: Union[str, Pattern]):
        super().__init__()
        if isinstance(regex, str):
            regex = re.compile(regex)
        self.regex = regex

    def _matches(self, other: str) -> bool:
        return re.search(self.regex, other) is not None

    def _description(self) -> str:
        return repr(self.regex)



from decimal import Decimal

from pychoir import ConvertsTo, IsInstance, OfType


def test_is_instance():
    assert OfType is IsInstance

    assert [1] == [IsInstance(int, str)]
    assert 'str' == IsInstance(int, str)
    assert {'a': 'abc'} == {'a': IsInstance(str)}
    assert not ['abc'] == [IsInstance(int)]

    assert str([IsInstance(int)]) == '[IsInstance(int)]'
    assert str(IsInstance(str, int)) == 'IsInstance(str, int)'


def test_converts_to():
    assert ['1'] == [ConvertsTo(int)]
    assert '1' == ConvertsTo(Decimal)
    assert 'asd' != ConvertsTo(int)
    assert [] != ConvertsTo(Decimal)

    assert str(ConvertsTo(int)) == 'ConvertsTo(int)'
    assert str(ConvertsTo(Decimal)) == 'ConvertsTo(Decimal)'



from contextlib import contextmanager
from typing import Iterator

from pychoir import (
    NE,
    AllOf,
    And,
    AnyOf,
    EqualTo,
    GreaterThan,
    HasLength,
    IsInstance,
    IsNoneOf,
    Not,
    Or,
    ResultsTrueFor,
)


def test_and():
    assert AllOf is And

    assert [1] == [And(IsInstance(int), GreaterThan(0))]
    assert {'a': [1]} == {'a': And(IsInstance(list), HasLength(GreaterThan(0)))}
    assert not {'a': []} == {'a': And(IsInstance(list), HasLength(GreaterThan(0)))}

    assert str(And(IsInstance(list), HasLength(GreaterThan(0)))) == 'And(IsInstance(list), HasLength(GreaterThan(0)))'


def test_or():
    assert AnyOf is Or

    assert [1] == [Or(0, 1)]
    assert {'a': [1]} == {'a': [Or(0, 1, 2)]}
    assert not [2] == [Or(0, 1)]

    assert str([Or(0, 1)]) == '[Or(0, 1)]'


def test_not():
    assert IsNoneOf is Not

    assert [1] == [IsNoneOf(0, 2, GreaterThan(3))]
    assert {'a': [None]} == {'a': IsNoneOf(Ellipsis, 0, [])}
    assert not [1] == [IsNoneOf(0, 1, 2)]

    assert str([Not(0, 1, 2)]) == '[Not(0, 1, 2)]'


def test_not_or():
    @contextmanager
    def make_not_or_123() -> Iterator[Not]:
        yield Not(Or(EqualTo(1), 2, 3))

    with make_not_or_123() as not_or_123:
        assert 4 == not_or_123
        assert str(not_or_123) == 'Not(Or(EqualTo(1), 2, 3))'

    with make_not_or_123() as not_or_123:
        assert not 4 != not_or_123
        assert str(not_or_123) == 'Not(Or(EqualTo(1)[FAILED for 4], 2, 3)[FAILED for 4])[FAILED for 4]'

    with make_not_or_123() as not_or_123:
        for i in (1, 2, 3):
            assert i != not_or_123
        assert str(not_or_123) == 'Not(Or(EqualTo(1), 2, 3))'

    with make_not_or_123() as not_or_123:
        for i in (1, 2, 3):
            assert not i == not_or_123
        assert str(not_or_123) == 'Not(Or(EqualTo(1)[FAILED for 1], 2, 3)[FAILED for (1, 2, 3)])[FAILED for (1, 2, 3)]'


def test_not_not():
    @contextmanager
    def make_not_one() -> Iterator[Not]:
        yield Not(1)

    with make_not_one() as not_one:
        assert 2 == not_one
        assert str(not_one) == 'Not(1)'

    with make_not_one() as not_one:
        assert 1 != not_one
        assert str(not_one) == 'Not(1)'

    with make_not_one() as not_one:
        assert not 2 != not_one
        assert str(not_one) == 'Not(1)[FAILED for 2]'

    with make_not_one() as not_one:
        assert not 1 == not_one
        assert str(not_one) == 'Not(1)[FAILED for 1]'

    @contextmanager
    def make_not_eq_one() -> Iterator[Not]:
        yield Not(EqualTo(1))

    with make_not_eq_one() as not_eq_one:
        assert 2 == not_eq_one
        assert str(not_eq_one) == 'Not(EqualTo(1))'

    with make_not_eq_one() as not_eq_one:
        assert 1 != not_eq_one
        assert str(not_eq_one) == 'Not(EqualTo(1))'

    with make_not_eq_one() as not_eq_one:
        assert not 2 != not_eq_one
        assert str(not_eq_one) == 'Not(EqualTo(1)[FAILED for 2])[FAILED for 2]'

    with make_not_eq_one() as not_eq_one:
        assert not 1 == not_eq_one
        assert str(not_eq_one) == 'Not(EqualTo(1)[FAILED for 1])[FAILED for 1]'

    @contextmanager
    def make_not_not_one() -> Iterator[Not]:
        yield Not(Not(1))

    with make_not_not_one() as not_not_one:
        assert 1 == not_not_one
        assert str(not_not_one) == 'Not(Not(1))'

    with make_not_not_one() as not_not_one:
        assert 2 != not_not_one
        assert str(not_not_one) == 'Not(Not(1))'

    with make_not_not_one() as not_not_one:
        assert not 1 != not_not_one
        assert str(not_not_one) == 'Not(Not(1)[FAILED for 1])[FAILED for 1]'

    with make_not_not_one() as not_not_one:
        assert not 2 == not_not_one
        assert str(not_not_one) == 'Not(Not(1)[FAILED for 2])[FAILED for 2]'

    @contextmanager
    def make_not_ne_one() -> Iterator[Not]:
        yield Not(NE(1))

    with make_not_ne_one() as not_ne_one:
        assert 1 == not_ne_one
        assert str(not_ne_one) == 'Not(NotEqualTo(1))'

    with make_not_ne_one() as not_ne_one:
        assert 2 != not_ne_one
        assert str(not_ne_one) == 'Not(NotEqualTo(1))'

    with make_not_ne_one() as not_ne_one:
        assert not 1 != not_ne_one
        assert str(not_ne_one) == 'Not(NotEqualTo(1)[FAILED for 1])[FAILED for 1]'

    with make_not_ne_one() as not_ne_one:
        assert not 2 == not_ne_one
        assert str(not_ne_one) == 'Not(NotEqualTo(1)[FAILED for 2])[FAILED for 2]'

    @contextmanager
    def make_not_not_eq_one() -> Iterator[Not]:
        yield Not(Not(EqualTo(1)))

    with make_not_not_eq_one() as not_not_eq_one:
        assert 1 == not_not_eq_one
        assert str(not_not_eq_one) == 'Not(Not(EqualTo(1)))'

    with make_not_not_eq_one() as not_not_eq_one:
        assert 2 != not_not_eq_one
        assert str(not_not_eq_one) == 'Not(Not(EqualTo(1)))'

    with make_not_not_eq_one() as not_not_eq_one:
        assert not 1 != not_not_eq_one
        assert str(not_not_eq_one) == 'Not(Not(EqualTo(1)[FAILED for 1])[FAILED for 1])[FAILED for 1]'

    with make_not_not_eq_one() as not_not_eq_one:
        assert not 2 == not_not_eq_one
        assert str(not_not_eq_one) == 'Not(Not(EqualTo(1)[FAILED for 2])[FAILED for 2])[FAILED for 2]'

    @contextmanager
    def make_not_not_eq_one_eq_two() -> Iterator[Not]:
        yield Not(Not(EqualTo(1), EqualTo(2)))

    with make_not_not_eq_one_eq_two() as not_not_eq_one_eq_two:
        assert 1 == not_not_eq_one_eq_two
        assert str(not_not_eq_one_eq_two) == 'Not(Not(EqualTo(1), EqualTo(2)))'

    with make_not_not_eq_one_eq_two() as not_not_eq_one_eq_two:
        assert not 2 != not_not_eq_one_eq_two
        assert (str(not_not_eq_one_eq_two)
                == 'Not(Not(EqualTo(1), EqualTo(2)[FAILED for 2])[FAILED for 2])[FAILED for 2]')

    with make_not_not_eq_one_eq_two() as not_not_eq_one_eq_two:
        assert not 1 != not_not_eq_one_eq_two
        assert (str(not_not_eq_one_eq_two)
                == 'Not(Not(EqualTo(1)[FAILED for 1], EqualTo(2))[FAILED for 1])[FAILED for 1]')

    with make_not_not_eq_one_eq_two() as not_not_eq_one_eq_two:
        assert 2 == not_not_eq_one_eq_two
        assert str(not_not_eq_one_eq_two) == 'Not(Not(EqualTo(1), EqualTo(2)))'


def test_results_true_for():
    assert {'a': 1} == {'a': ResultsTrueFor(bool)}
    assert ['foobar'] == [ResultsTrueFor(lambda x: x.startswith('foo'))]
    assert not {'a': 0} == {'a': ResultsTrueFor(bool)}

    assert str({'a': ResultsTrueFor(bool)}) == "{'a': ResultsTrueFor(<class 'bool'>)}"



from typing import Any, Optional, TypeVar

from pychoir import M, Matches

T = TypeVar('T')


def test_hamcrest() -> None:
    class Description:
        pass

    class TestHamcrestLike:
        def __init__(self, return_value: bool):
            self.return_value = return_value

        def matches(self, item: T, mismatch_description: Optional[Description] = None) -> bool:
            return self.return_value

    matching_matcher = TestHamcrestLike(True)
    mismatching_matcher = TestHamcrestLike(False)

    assert ['a'] == [Matches(matching_matcher)]
    assert not ['a'] == [M(mismatching_matcher)]

    assert (str(Matches(TestHamcrestLike(True)))
            .startswith('Matches(<tests.test_integration.test_hamcrest.<locals>.TestHamcrestLike object at'))


def test_matches() -> None:
    assert M is Matches

    class TestMatcherLike:
        def __init__(self, return_value: bool):
            self.return_value = return_value

        def matches(self, _: Any) -> bool:
            return self.return_value

    matching_matcher = TestMatcherLike(True)
    mismatching_matcher = TestMatcherLike(False)

    assert ['a'] == [Matches(matching_matcher)]
    assert not ['a'] == [Matches(mismatching_matcher)]



from typing import Any

from pychoir import Matcher


class IsOdd(Matcher):
    """A Matcher checking that the value compared against it is odd.

    Usage:
      >>> from pychoir import IsOdd
      >>> 5 == IsOdd()
      True
      >>> 4 == IsOdd()
      False
    """
    def _matches(self, other: Any) -> bool:
        return bool(other % 2 == 1)

    def _description(self) -> str:
        return ''


class IsEven(Matcher):
    """A Matcher checking that the value compared against it is even.

    Usage:
      >>> from pychoir import IsEven
      >>> 4 == IsEven()
      True
      >>> 5 == IsEven()
      False
    """
    def _matches(self, other: Any) -> bool:
        return bool(other % 2 == 0)

    def _description(self) -> str:
        return ''


class IsPositive(Matcher):
    """A Matcher checking that the value compared against it is positive.

    Usage:
      >>> from pychoir import IsPositive
      >>> 1 == IsPositive()
      True
      >>> 0 == IsPositive()
      False
    """
    def _matches(self, other: Any) -> bool:
        return bool(other > 0)

    def _description(self) -> str:
        return ''


class IsNonNegative(Matcher):
    """A Matcher checking that the value compared against it is non-negative.

    Usage:
      >>> from pychoir import IsPositive
      >>> 1 == IsNonNegative()
      True
      >>> 0 == IsNonNegative()
      True
      >>> -1 == IsNonNegative()
      False
    """
    def _matches(self, other: Any) -> bool:
        return bool(other >= 0)

    def _description(self) -> str:
        return ''


class IsNegative(Matcher):
    """A Matcher checking that the value compared against it is negative.

    Usage:
      >>> from pychoir import IsNegative
      >>> -1 == IsNegative()
      True
      >>> 0 == IsNegative()
      False
    """
    def _matches(self, other: Any) -> bool:
        return bool(other < 0)

    def _description(self) -> str:
        return ''



from typing import Any, Callable, Optional, Type

from pychoir import Matcher


class WhenPassedTo:
    """Matchers that check how a value behaves when passed to a callable.

    :param callable_: The callable to pass the value to.

    Usage:
      >>> from pychoir import WhenPassedTo
      >>> "foo" == WhenPassedTo(int).raises(ValueError)
      True
      >>> "5" == WhenPassedTo(int).does_not_raise()
      True
      >>> "5" == WhenPassedTo(int).returns(5)
      True
    """
    def __init__(self, callable_: Callable[[Any], Any]):
        self.callable = callable_

    def returns(self, value: Any) -> Matcher:
        return _Returns(self.callable, value)

    def raises(self, exception: Optional[Type[Exception]] = None) -> Matcher:
        return _Raises(self.callable, exception)

    def does_not_raise(self) -> Matcher:
        return _DoesNotRaise(self.callable)


class _Returns(Matcher):
    def __init__(self, callable_: Callable[[Any], Any], value: Any):
        super().__init__(name='WhenPassedTo')
        self.callable = callable_
        self.value = value

    def _matches(self, other: Any) -> bool:
        return bool(self.callable(other) == self.value)

    def _description(self) -> str:
        callable_name = _name_or_repr(self.callable)

        return f'{callable_name}).returns({self.value!r}'


class _Raises(Matcher):
    def __init__(self, callable_: Callable[..., Any], exception: Optional[Type[Exception]]):
        super().__init__(name='WhenPassedTo')
        self.callable = callable_
        self.exception = exception

    def _matches(self, other: Any) -> bool:
        raised = None
        try:
            self.callable(other)
        except Exception as e:
            raised = e
        if self.exception is not None:
            return isinstance(raised, self.exception)
        else:
            return raised is not None

    def _description(self) -> str:
        callable_name = _name_or_repr(self.callable)
        exception_name = _name_or_repr(self.exception)

        return f'{callable_name}).raises({exception_name}'


class _DoesNotRaise(Matcher):
    def __init__(self, callable_: Callable[..., Any]):
        super().__init__(name='WhenPassedTo')
        self.callable = callable_

    def _matches(self, other: Any) -> bool:
        try:
            self.callable(other)
        except Exception:
            return False
        return True

    def _description(self) -> str:
        callable_name = _name_or_repr(self.callable)
        return f'{callable_name}).does_not_raise('


def _name_or_repr(thing: Any) -> str:
    return getattr(thing, '__name__', None) or repr(thing)



import sys
from contextlib import contextmanager
from typing import Any, Dict, List, cast
from unittest.mock import MagicMock

import pytest

from pychoir import (
    And,
    Anything,
    EqualTo,
    GreaterThan,
    InAnyOrder,
    IsInstance,
    IsOdd,
    IsTruthy,
    Matchable,
    Matcher,
    that,
)


def test_matchable():
    def matcher(matchable: Matchable) -> bool:
        return matchable == 5

    assert matcher(5) is True


class TestMatcher:
    class _TestMatcher(Matcher):
        def __init__(self, does_match: bool):
            super().__init__(name='Matcher')
            self.does_match = does_match

        def _matches(self, other: Any) -> bool:
            return self.does_match

        def _description(self) -> str:
            return f'{self.does_match}'

    def test_as(self):
        instance = self._TestMatcher(True)
        assert instance.as_(int) is cast(int, instance)
        assert type(instance.as_(int)) is cast(int, self._TestMatcher)
        assert instance.as_(int) == 5

    def test_eq(self):
        assert self._TestMatcher(True) == 1
        assert 0 == self._TestMatcher(True)
        assert not self._TestMatcher(False) == 0
        assert not 0 == self._TestMatcher(False)

    def test_ne(self):
        assert 0 != self._TestMatcher(False)

    def test_str(self):
        assert str(self._TestMatcher(True)) == 'Matcher(True)'

    def test_repr(self):
        assert repr(self._TestMatcher(True)) == 'Matcher(True)'

    def test_failure_report(self):
        with pytest.raises(AssertionError) as exc_info:
            assert 'foo' == self._TestMatcher(False)
        assert "Matcher(False)[FAILED for 'foo']" in str(exc_info.value)

    def test_failure_report_on_mismatch_outside_matcher(self):
        with pytest.raises(AssertionError) as exc_info:
            matching_matcher = self._TestMatcher(True)
            assert {'a': 1, 'b': 0} == {'a': matching_matcher, 'b': 1}
        assert "assert {'a': 1, 'b': 0} == {'a': Matcher(True), 'b': 1}" in str(exc_info.value)
        assert str(matching_matcher) == 'Matcher(True)'

    def test_nested_match(self):
        @contextmanager
        def new_matching_matcher():
            yield self._TestMatcher(True)
        with new_matching_matcher() as matching_matcher:
            assert str(matching_matcher) == 'Matcher(True)'
        with new_matching_matcher() as matching_matcher:
            assert Anything().nested_match(matching_matcher, 1)
            assert str(matching_matcher) == 'Matcher(True)'
        with new_matching_matcher() as matching_matcher:
            assert Anything().nested_match(matching_matcher, 1, expect_mismatch=True)
            assert str(matching_matcher) == 'Matcher(True)[FAILED for 1]'
            assert Anything().nested_match(matching_matcher, 2, expect_mismatch=True)
            assert str(matching_matcher) == 'Matcher(True)[FAILED for (1, 2)]'

        @contextmanager
        def new_mismatching_matcher():
            yield self._TestMatcher(False)
        with new_mismatching_matcher() as mismatching_matcher:
            assert str(mismatching_matcher) == 'Matcher(False)'
        with new_mismatching_matcher() as mismatching_matcher:
            assert not Anything().nested_match(mismatching_matcher, 3, expect_mismatch=True)
            assert str(mismatching_matcher) == 'Matcher(False)'
        with new_mismatching_matcher() as mismatching_matcher:
            assert not Anything().nested_match(mismatching_matcher, 3)
            assert str(mismatching_matcher) == 'Matcher(False)[FAILED for 3]'
            assert not Anything().nested_match(mismatching_matcher, 4)
            assert str(mismatching_matcher) == 'Matcher(False)[FAILED for (3, 4)]'

        assert Anything().nested_match(1, 1)
        assert Anything().nested_match(1, 1, expect_mismatch=True)
        assert not Anything().nested_match(1, 2)
        assert not Anything().nested_match(1, 2, expect_mismatch=True)

    def test_and_operator(self) -> None:
        assert 5 == IsInstance(int) & 5
        assert 5 != EqualTo(5) & IsInstance(float)

        assert str(EqualTo(5) & IsInstance(int)) == '(EqualTo(5) & IsInstance(int))'

        failing_and_operator = EqualTo(5) & IsInstance(float)
        assert not 5 == failing_and_operator
        assert str(failing_and_operator) == '(EqualTo(5) & IsInstance(float)[FAILED for 5])[FAILED for 5]'

        one_and = EqualTo(5) & IsInstance(int)
        another_and = IsOdd() & IsTruthy()
        combo_and = one_and & another_and
        assert 5 == combo_and
        assert 6 != combo_and
        assert str(combo_and) == '(EqualTo(5) & IsInstance(int) & IsOdd() & IsTruthy())'

        and_and_normal = one_and & IsTruthy()
        assert 5 == and_and_normal
        assert str(and_and_normal) == '(EqualTo(5) & IsInstance(int) & IsTruthy())'

    def test_or_operator(self) -> None:
        assert 5 == IsInstance(float) | 5
        assert 5 != EqualTo(6) | IsInstance(float)

        assert str(EqualTo(5) | IsInstance(int)) == '(EqualTo(5) | IsInstance(int))'

        failing_or_operator = EqualTo(5) | IsInstance(float)
        assert not 6 == failing_or_operator
        assert str(failing_or_operator) == '(EqualTo(5)[FAILED for 6] | IsInstance(float)[FAILED for 6])[FAILED for 6]'

        one_or = EqualTo(5) | IsInstance(int)
        another_or = IsOdd() | IsTruthy()
        combo_or = one_or | another_or
        assert 5 == combo_or
        assert 6 == combo_or
        assert str(combo_or) == '(EqualTo(5) | IsInstance(int) | IsOdd() | IsTruthy())'

        or_or_normal = one_or | IsTruthy()
        assert 6 == or_or_normal
        assert str(or_or_normal) == '(EqualTo(5) | IsInstance(int) | IsTruthy())'


def test_matcher_in_mock_call_params():
    m = MagicMock()

    m(5)
    m.assert_called_once_with(Anything())
    m.assert_called_once_with(And(IsInstance(int), GreaterThan(3)))

    with pytest.raises(AssertionError) as exc_info:
        m.assert_called_once_with(GreaterThan(5))
    if sys.version_info >= (3, 8):
        assert ("\n".join(line.lstrip() for line in str(exc_info.value).split("\n"))
                == "expected call not found.\n"
                   "Expected: mock(GreaterThan(5)[FAILED for 5])\n"
                   "Actual: mock(5)")
    else:
        assert (str(exc_info.value)
                == "Expected call: mock(GreaterThan(5)[FAILED for 5])\n"
                   "Actual call: mock(5)")

    m.do_stuff_to([{'a': 1}, {'a': 2}])
    m.do_stuff_to.assert_called_once_with(InAnyOrder([{'a': 2}, {'a': 1}]))


def test_assert_that_matches():
    assert that(5).matches(And(EqualTo(5), IsInstance(int)))

    with pytest.raises(AssertionError) as exc_info:
        assert that(5).matches(EqualTo(4))
    assert str(exc_info.value).split('\n')[0] == 'assert that(5).matches(EqualTo(4)[FAILED for 5])'


if sys.version_info >= (3, 7):
    from dataclasses import dataclass

    def test_matcher_in_dataclass():
        @dataclass
        class Data:
            i: int
            ldsb: List[Dict[str, bool]]

        assert Data(1, [{'foo': True}]) == Data(IsInstance(int).as_(int), [{'foo': IsTruthy().as_(bool)}])

        assert (str(Data(IsInstance(int).as_(int), [{'foo': IsTruthy().as_(bool)}]))
                == "test_matcher_in_dataclass.<locals>.Data(i=IsInstance(int), ldsb=[{'foo': IsTruthy()}])")



from pychoir import (
    All,
    Anything,
    EqualTo,
    In,
    Is,
    IsFalsy,
    IsNoneOr,
    IsTruthy,
    LessThan,
    Not,
    OneOf,
    Optionally,
)


def test_anything():
    assert [1] == [Anything()]
    assert [None] == [Anything()]
    assert [Ellipsis] == [Anything()]
    assert not [] == [Anything()]

    assert str(Anything()) == 'Anything()'


def test_is():
    assert [None] == [Is(None)]
    assert not [{}] == [Is({})]
    assert not [] == [Is(None)]

    assert str([Is(None)]) == '[Is(None)]'


def test_is_none_or():
    assert Optionally is IsNoneOr
    assert [None, 1] == All(Optionally(1))
    assert [None, 1] == [IsNoneOr(1), IsNoneOr(LessThan(0), EqualTo(1))]
    assert not [2] == [IsNoneOr(1)]

    assert str([IsNoneOr('1')]) == "[IsNoneOr('1')]"


def test_is_truthy():
    assert [1, True, [0], {0}, {'': None}, (0,)] == All(IsTruthy())
    assert [0, False, [], set(), {}, ()] == All(Not(IsTruthy()))

    assert str(All(Not(IsTruthy()))) == 'All(Not(IsTruthy()))'


def test_is_falsy():
    assert [0, False, [], set(), {}, ()] == All(IsFalsy())
    assert [1, True, [0], {0}, {'': None}, (0,)] == All(Not(IsFalsy()))

    assert str(All(Not(IsFalsy()))) == 'All(Not(IsFalsy()))'


def test_in():
    assert OneOf is In

    assert [1] == [In([1, 2])]
    assert 'a' == In({'a': 5})
    assert not [0] == [In([1, 2])]

    assert str([In(['1', '2'])]) == "[In(['1', '2'])]"



import re

from pychoir import EndsWith, MatchesRegex, StartsWith


def test_startswith():
    assert 'foobar' == StartsWith('foo')
    assert 'barbar' != StartsWith('foo')

    assert str(StartsWith('foo')) == "StartsWith('foo')"


def test_endswith():
    assert 'foobar' == EndsWith('bar')
    assert 'foo' != EndsWith('bar')

    assert str(EndsWith('bar')) == "EndsWith('bar')"


def test_matches_regex():
    assert 'foobar' == MatchesRegex(r'^foo')
    assert 'foobar' == MatchesRegex(re.compile(r'bar$', flags=re.VERBOSE))
    assert 'bafoo' != MatchesRegex(r'^foo')

    assert str(MatchesRegex(re.compile(r'^foo', re.VERBOSE))) == "MatchesRegex(re.compile('^foo', re.VERBOSE))"



from pychoir import (
    EQ,
    GT,
    GTE,
    LT,
    LTE,
    NE,
    EqualTo,
    GreaterThan,
    GreaterThanOrEqualTo,
    LessThan,
    LessThanOrEqualTo,
    NotEqualTo,
)


def test_equal_to():
    assert EQ is EqualTo

    assert [1] == [EqualTo(1)]
    assert {'a': 1} == {'a': EqualTo(True)}
    assert not [0] == [EqualTo(1)]

    assert str(EqualTo(1)) == 'EqualTo(1)'


def test_not_equal_to():
    assert NE is NotEqualTo

    assert [0] == [NotEqualTo(1)]
    assert {'a': 0} == {'a': NotEqualTo(1)}
    assert not [0] == [NotEqualTo(0)]

    assert str(NotEqualTo(1)) == 'NotEqualTo(1)'


def test_greater_than():
    assert GT is GreaterThan

    assert [0] == [GreaterThan(-1)]
    assert {'a': 0} == {'a': GreaterThan(-1)}
    assert not [0] == [GreaterThan(0)]

    assert str(GreaterThan(-1)) == 'GreaterThan(-1)'


def test_greater_than_or_equal_to():
    assert GTE is GreaterThanOrEqualTo

    assert [0, 1] == [GreaterThanOrEqualTo(0), GreaterThanOrEqualTo(0)]
    assert {'a': 1} == {'a': GreaterThanOrEqualTo(1)}
    assert not [-1] == [GreaterThanOrEqualTo(0)]

    assert str(GreaterThanOrEqualTo(0)) == 'GreaterThanOrEqualTo(0)'


def test_lesser_than():
    assert LT is LessThan

    assert [0] == [LessThan(1)]
    assert {'a': 0} == {'a': LessThan(1)}
    assert not [0] == [LessThan(0)]

    assert str(LessThan(2)) == 'LessThan(2)'


def test_lesser_than_or_equal_to():
    assert LTE is LessThanOrEqualTo

    assert [0, 1] == [LessThanOrEqualTo(1), LessThanOrEqualTo(1)]
    assert {'a': 0} == {'a': LessThanOrEqualTo(0)}
    assert not [0] == [LessThanOrEqualTo(-1)]

    assert str(LessThanOrEqualTo(-1)) == 'LessThanOrEqualTo(-1)'



from typing import Any, Iterable

from pychoir.core import Matchable, Matcher


class Anything(Matcher):
    """A Matcher that matches anything.

    Usage:
      >>> from pychoir import Anything
      >>> None == Anything()
      True
      >>> {'a': None} == {'a': Anything()}
      True
      >>> {'b': 1} == {'a': Anything()}
      False
    """
    def _matches(self, _: Any) -> bool:
        return True

    def _description(self) -> str:
        return ''


class Is(Matcher):
    """A Matcher checking that the compared value :code:`is` the passed value.

    In Python :code:`is` is used to check whether two objects are the same object in memory.

    :param value: The value to compare against.

    Usage:
      >>> from pychoir import Is
      >>> None == Is(None)
      True
      >>> [] == Is([])
      False
    """
    def __init__(self, value: Any):
        super().__init__()
        self.value = value

    def _matches(self, other: Any) -> bool:
        return other is self.value

    def _description(self) -> str:
        return repr(self.value)


class IsNoneOr(Matcher):
    """A Matcher checking that the compared value :code:`is None` or equal to the passed value.

    This is useful for checking :code:`Optional` values.

    :param matchers: The value(s) and/or matcher(s) to compare against.

    Usage:
      >>> from pychoir import IsNoneOr
      >>> None == IsNoneOr(5)
      True
      >>> 5 == IsNoneOr(5)
      True
      >>> 4 == IsNoneOr(5)
      False
    """
    def __init__(self, *matchers: Matchable):
        super().__init__()
        self.matchers = matchers

    def _matches(self, other: Any) -> bool:
        return other is None or any(self.nested_match(matcher, other) for matcher in self.matchers)

    def _description(self) -> str:
        return ', '.join(map(repr, self.matchers))


Optionally = IsNoneOr


class IsTruthy(Matcher):
    """A Matcher checking that the compared value is *truthy* when used as a condition.

    In other words, checks whether :code:`if value: ... else: ...` would go to the if branch.

    Usage:
      >>> from pychoir import IsTruthy
      >>> '' == IsTruthy()
      False
      >>> 5 == IsTruthy()
      True
    """
    def _matches(self, other: Any) -> bool:
        return bool(other)

    def _description(self) -> str:
        return ''


class IsFalsy(Matcher):
    """A Matcher checking that the compared value is *falsy* when used as a condition.

    In other words, checks whether :code:`if value: ... else: ...` would go to the else branch.

    Usage:
      >>> from pychoir import IsFalsy
      >>> '' == IsFalsy()
      True
      >>> 5 == IsFalsy()
      False
    """
    def _matches(self, other: Any) -> bool:
        return not bool(other)

    def _description(self) -> str:
        return ''


class In(Matcher):
    """A Matcher checking that the compared value is in the passed iterable.

    Usage:
      >>> from string import ascii_lowercase
      >>> from pychoir import In
      >>> 'a' == In(ascii_lowercase)
      True
      >>> 'A' == In(ascii_lowercase)
      False
    """
    def __init__(self, allowed_values: Iterable[Any]):
        super().__init__()
        self.allowed_values = allowed_values

    def _matches(self, other: Any) -> bool:
        return other in self.allowed_values

    def _description(self) -> str:
        return repr(self.allowed_values)


OneOf = In



"""
Useful extensions for standard library
"""
from enum import Enum
from typing import Any, Sequence, Tuple, TypeVar, Union

T = TypeVar('T')


def sequence_or_its_only_member(sequence: Sequence[T]) -> Union[Sequence[T], T]:
    if len(sequence) == 1:
        return sequence[0]
    else:
        return sequence


def i_removed(tuple_: Tuple[Any, ...], index: Any) -> Tuple[Any, ...]:
    return tuple_[:index] + tuple_[index + 1:]


class DefaultType(Enum):
    DEFAULT = 'default'


Default = DefaultType.DEFAULT



from typing import Any

from pychoir.callables import WhenPassedTo


def test_returns() -> None:
    assert "5" == WhenPassedTo(int).returns(5)
    assert "5" != WhenPassedTo(int).returns(6)

    assert str(WhenPassedTo(int).returns(5)) == 'WhenPassedTo(int).returns(5)'


def test_raises() -> None:
    def raiser(_: Any) -> None:
        raise RuntimeError('raiser raised')

    assert "lol" == WhenPassedTo(raiser).raises()
    assert "lol" == WhenPassedTo(int).raises(ValueError)
    assert 5 != WhenPassedTo(int).raises()

    assert str(WhenPassedTo(raiser).raises(ValueError)) == 'WhenPassedTo(raiser).raises(ValueError)'


def test_does_not_raise() -> None:
    def raiser(_: Any) -> None:
        raise RuntimeError('raiser raised')

    assert "5" == WhenPassedTo(int).does_not_raise()
    assert "lol" != WhenPassedTo(int).does_not_raise()
    assert not Any == WhenPassedTo(raiser).does_not_raise()

    assert str(WhenPassedTo(raiser).does_not_raise()) == 'WhenPassedTo(raiser).does_not_raise()'



import sys
from typing import Any, TypeVar

from pychoir.core import Matcher

T = TypeVar('T', contravariant=True)

if sys.version_info >= (3, 8):
    from typing import Protocol

    class _MatcherLike(Protocol[T]):
        def matches(self, other: T) -> bool:
            ...  # pragma: no cover

    MatcherLike = _MatcherLike[T]
else:
    MatcherLike = Any


class Matches(Matcher):
    def __init__(self, *matchers: MatcherLike):
        super().__init__()
        self.matchers = matchers

    def _matches(self, other: T) -> bool:
        return any(matcher.matches(other) for matcher in self.matchers)

    def _description(self) -> str:
        return ', '.join(map(repr, self.matchers))


M = Matches



import sys
from typing import Any, Iterable, Mapping, Optional, Sequence, Tuple, Union

from pychoir.core import Matchable, Matcher, Transformer
from pychoir.utils import i_removed

if sys.version_info >= (3, 8):
    from typing import Protocol

    class Lengthy(Protocol):
        def __len__(self) -> int:
            ...  # pragma: no cover
else:
    Lengthy = Any


class IsEmpty(Matcher):
    """A Matcher checking that the `len()` of the compared value is 0.

    Usage:
      >>> from pychoir import All, IsEmpty, Not
      >>> ('', [], {}, set(), tuple()) == All(IsEmpty())
      True
      >>> {'not': 'empty'} == Not(IsEmpty())
      True
    """

    def _matches(self, other: Lengthy) -> bool:
        return len(other) == 0

    def _description(self) -> str:
        return ''


class HasLength(Matcher):
    """A Matcher checking that the `len()` of the compared value matches the passed Matchable.

    :param matcher: The value or Matcher to compare against.

    Usage:
      >>> from pychoir import GreaterThan, HasLength
      >>> 'foo' == HasLength(3)
      True
      >>> 'foo' == HasLength(GreaterThan(2))
      True
    """
    def __init__(self, matcher: Matchable):
        super().__init__()
        self.matcher = matcher

    def _matches(self, other: Lengthy) -> bool:
        return self.nested_match(self.matcher, len(other))

    def _description(self) -> str:
        return repr(self.matcher)


Len = HasLength


class All(Matcher):
    """A Matcher checking that all values in a container match passed Matchables.

    :param matchers: The value(s) and/or Matcher(s) to compare against.

    Usage:
      >>> from pychoir import All, IsInstance
      >>> 'aaa' == All('a')
      True
      >>> [1, 2, 3] == All(IsInstance(int))
      True
    """
    def __init__(self, *matchers: Matchable):
        super().__init__()
        self.matchers = matchers

    def _matches(self, iterable: Iterable[Any]) -> bool:
        return all(self.nested_match(matcher, value) for value in iterable for matcher in self.matchers)

    def _description(self) -> str:
        return ', '.join(map(repr, self.matchers))


class AreNot(Matcher):
    """A Matcher checking that none of the values in a container match passed Matchables.

    :param matchers: The value(s) and/or Matcher(s) to compare against.

    Usage:
      >>> from pychoir import AreNot, IsInstance
      >>> 'abc' == AreNot('a', 'b')
      False
      >>> [1, 2, 3] == AreNot(IsInstance(str))
      True
    """
    def __init__(self, *matchers: Matchable):
        super().__init__()
        self.matchers = matchers

    def _matches(self, iterable: Iterable[Any]) -> bool:
        return not any(self.nested_match(matcher, value, expect_mismatch=True)
                       for value in iterable for matcher in self.matchers)

    def _description(self) -> str:
        return ', '.join(map(repr, self.matchers))


class Contains(Matcher):
    """A Matcher checking that a container contains the passed value.

    :param value: The value to find in the container.

    Usage:
      >>> from pychoir import Contains
      >>> 'abc' == Contains('a')
      True
      >>> [1, 2, 3] == Contains(4)
      False
    """
    def __init__(self, value: Any):
        super().__init__()
        self.value = value

    def _matches(self, other: Any) -> bool:
        return self.value in other

    def _description(self) -> str:
        return repr(self.value)


class ContainsAllOf(Matcher):
    """A Matcher checking that a container contains *at least* the passed values.

    Plural of :class:`Contains`.

    :param values: The value(s) to find in the container.

    Usage:
      >>> from pychoir import ContainsAllOf
      >>> 'abc' == ContainsAllOf('a', 'b')
      True
      >>> [1, 2, 3] == ContainsAllOf(3, 4)
      False
    """
    def __init__(self, *values: Any):
        super().__init__()
        self.values = values

    def _matches(self, other: Any) -> bool:
        return all(value in other for value in self.values)

    def _description(self) -> str:
        return ', '.join(map(repr, self.values))


class ContainsAnyOf(Matcher):
    """A Matcher checking that a container contains *at least one of* the passed values.

    :param values: The value(s) to find in the container.

    Usage:
      >>> from pychoir import ContainsAnyOf
      >>> 'abc' == ContainsAnyOf('a', 'b')
      True
      >>> [1, 2, 3] == ContainsAnyOf(3, 4)
      True
    """
    def __init__(self, *values: Any):
        super().__init__()
        self.values = values

    def _matches(self, other: Any) -> bool:
        return any(value in other for value in self.values)

    def _description(self) -> str:
        return ', '.join(map(repr, self.values))


class ContainsNoneOf(Matcher):
    """A Matcher checking that a container contains none of the passed values.

    :param values: The value(s) to find in the container.

    Usage:
      >>> from pychoir import ContainsNoneOf
      >>> 'abc' == ContainsNoneOf('a', 'b')
      False
      >>> [1, 2, 3] == ContainsNoneOf(4)
      True
    """
    def __init__(self, *values: Any):
        super().__init__()
        self.values = values

    def _matches(self, other: Any) -> bool:
        return not any(value in other for value in self.values)

    def _description(self) -> str:
        return ', '.join(map(repr, self.values))


class DictContainsAllOf(Matcher):
    """A Matcher checking that a Mapping contains *at least* the passed Mapping.
    Usually this means that the passed dict is a subset of the one compared against.
    Keys expected to be absent can be set as :class:`NotPresent`.

    :param value: The Mapping to find in the Mapping compared against.

    Usage:
      >>> from pychoir import DictContainsAllOf, NotPresent
      >>> {'a': 1, 'b': 2, 'd': 3} == DictContainsAllOf({'a': 1, 'c': NotPresent})
      True
      >>> {'a': 1, 'c': 2, 'd': 3} == DictContainsAllOf({'a': 1, 'c': NotPresent})
      False
    """
    def __init__(self, value: Mapping[Any, Any]):
        super().__init__()
        self.expected = value

    def _matches(self, other: Mapping[Any, Any]) -> bool:
        match_dict = {key: value for key, value in other.items() if key in self.expected}
        for key in self.expected:
            if key not in match_dict:
                match_dict[key] = NotPresent
        return self.expected == match_dict

    def _description(self) -> str:
        return repr(self.expected)


class InAnyOrder(Matcher):
    """A Matcher checking that an Iterable contains *exactly* the passed items, in any order.

    The Iterable can be for example a list, tuple or set. Items can be Matchers and do not need to be hashable.

    :param values: An Iterable containing the expected items, in any order.

    Usage:
      >>> from pychoir import InAnyOrder, IsEven, IsOdd
      >>> [1, 2, 3, 3] == InAnyOrder([3, 2, 3, 1])
      True
      >>> [1, 2, 3] == InAnyOrder([3, 3, 2, 1])  # missing one 3
      False
      >>> [1, 2, 3, 3] == InAnyOrder([3, 2, 1])  # one 3 too many
      False
      >>> [1, 2, 3] == InAnyOrder((1, 2, 3))  # type of the Iterable does not matter
      True
      >>> [{'a': 1}, {'b': 2}] == InAnyOrder([{'b': IsEven()}, {'a': IsOdd()}])
      True
    """
    def __init__(self, values: Iterable[Matchable]):
        super().__init__()
        self.expected_values = values

    def _matches(self, other: Iterable[Any]) -> bool:
        return self.__has_a_solution(tuple(self.expected_values), tuple(other))

    def __has_a_solution(self, expected_left: Tuple[Matchable, ...], values_left: Tuple[Any, ...]) -> bool:
        if not expected_left and not values_left:
            return True  # expectations and values matched one-to-one
        if not expected_left or not values_left:
            return False  # expectations or values left but nothing to match with

        expected_head, expected_tail = expected_left[0], expected_left[1:]
        return any(self.nested_match(expected_head, value) and
                   self.__has_a_solution(expected_tail, i_removed(values_left, i))
                   for i, value in enumerate(values_left))

    def _description(self) -> str:
        return repr(self.expected_values)


class SetEquals(Matcher):
    """A Matcher checking that an Iterable has the expected items, duplicates ignored.
    Faster than :class:`InAnyOrder` but less pedantic about duplicates and requires hashable items.

    The Iterable can be for example a list, tuple or set. Items must be hashable.

    :param values: An Iterable containing the expected items, in any order.

    Usage:
      >>> from pychoir import SetEquals
      >>> [1, 2, 3, 3] == SetEquals([3, 2, 1])
      True
      >>> [1, 2] == SetEquals([3, 2, 1])
      False
    """
    def __init__(self, values: Iterable[Any]):
        super().__init__()
        self.expected_values = values

    def _matches(self, other: Iterable[Any]) -> bool:
        return set(self.expected_values) == set(other)

    def _description(self) -> str:
        return repr(self.expected_values)


class _NotPresent:
    def __str__(self) -> str:
        return 'NotPresent'

    def __repr__(self) -> str:
        return str(self)


NotPresent = _NotPresent()


class IsNotPresentOr(Matcher):
    """A Matcher checking that a value is either :class:`NotPresent` or matches the passed matcher.
    Usually used for example with :class:`DictContainsAllOf`.

    :param matcher: The Mapping to find in the Mapping compared against.

    Usage:
      >>> from pychoir import DictContainsAllOf, IsNotPresentOr
      >>> {'a': 2} == DictContainsAllOf({'a': IsNotPresentOr(2)})
      True
      >>> {} == DictContainsAllOf({'a': IsNotPresentOr(2)})
      True
    """
    def __init__(self, matcher: Matchable):
        super().__init__()
        self.matcher = matcher

    def _matches(self, other: Any) -> bool:
        return other is NotPresent or self.nested_match(self.matcher, other)

    def _description(self) -> str:
        return repr(self.matcher)


class _First(Matcher):
    def __init__(self, how_many: int, matcher: Matchable):
        super().__init__(name='First')
        self.how_many = how_many
        self.matcher = matcher

    def _matches(self, other: Sequence[Any]) -> bool:
        if self.how_many < 2:
            first = other[0]
        else:
            first = other[:self.how_many]
        return self.nested_match(self.matcher, first)

    def _description(self) -> str:
        matcher_string = repr(self.matcher)
        return f'{self.how_many})({matcher_string}'


class First(Transformer):
    """A Higher Order Matcher checking that the first `how_many` values match given matchers.
    The matcher is applied to the slice extracted from the passed :class:`Sequence`.

    :param how_many: The number of values to slice from the start of the sequence being compared.

    Usage:
      >>> from pychoir import First, All, GreaterThan
      >>> (1, 2, 3) == First(2)((1, 2))
      True
      >>> (1, 2, 3) == First()(1)
      True
    """
    def __init__(self, how_many: int = 1):
        super().__init__()
        self.how_many = how_many
        self.matcher: Optional[Matchable] = None

    def matches(self, matcher: Matchable) -> _First:
        """
        :param matcher: The :class:`Matcher` to apply on the extracted slice.
        """
        return _First(self.how_many, matcher)


class _Last(Matcher):
    def __init__(self, how_many: int, matcher: Matchable):
        super().__init__(name='Last')
        self.how_many = how_many
        self.matcher = matcher

    def _matches(self, other: Sequence[Any]) -> bool:
        if self.how_many < 2:
            last = other[-1]
        else:
            last = other[-self.how_many:]
        return self.nested_match(self.matcher, last)

    def _description(self) -> str:
        matcher_string = repr(self.matcher)
        return f'{self.how_many})({matcher_string}'


class Last(Transformer):
    """A Higher Order Matcher checking that the last `how_many` values match given matchers.
    The matcher is applied to the slice extracted from the passed :class:`Sequence`.

    :param how_many: The number of values to slice from the end of the sequence being compared.

    Usage:
      >>> from pychoir import Last, All, LessThan
      >>> (1, 2, 3) == Last(2)((2, 3))
      True
      >>> (1, 2, 3) == Last()(3)
      True
    """
    def __init__(self, how_many: int = 1):
        self.how_many = how_many

    def matches(self, matcher: Matchable) -> _Last:
        """
        :param matcher: The :class:`Matcher` to apply on the extracted slice.
        """
        return _Last(self.how_many, matcher)


class _Slice(Matcher):
    def __init__(self, slice_: Union[int, slice], matcher: Matchable) -> None:
        super().__init__(name=f'Slice[{slice_}]')
        self.slice = slice_
        self.matcher = matcher

    def _matches(self, other: Any) -> bool:
        sliced = other[self.slice]
        return self.nested_match(self.matcher, sliced)

    def _description(self) -> str:
        return repr(self.matcher)


class _SliceTransformer(Transformer):
    """Extracts `slice_` from the matched value before checking it with a Matcher.
    """
    def __init__(self, slice_: Union[int, slice]) -> None:
        self.slice: Union[int, slice] = slice_

    def matches(self, matcher: Matchable) -> _Slice:
        """
        :param matcher: The :class:`Matcher` to apply on the extracted slice.
        """
        return _Slice(self.slice, matcher)


class _SliceFactory:
    """A Higher Order Matcher checking that the sliced values match given matchers.
    The matcher is applied to the slice extracted from the passed :class:`Sequence`.

    Usage:
      >>> from pychoir import Slice, And, IsInstance
      >>> (1, 2, 3) == Slice[1:3]((2, 3))
      True
      >>> (1, 2, 3) == Slice[2](And(IsInstance(int), 3))
      True
    """
    def __getitem__(self, slice_: Union[int, slice]) -> _SliceTransformer:
        """
        :param slice_: The :class:`slice` to slice from the matched object.
        """
        return _SliceTransformer(slice_)


Slice = _SliceFactory()



from typing import Any, Type

from pychoir.core import Matcher


class IsInstance(Matcher):
    """A Matcher checking that the compared object's type matches the passed one.

    :param types: The expected type(s) of compared values.

    Usage:
      >>> from pychoir import IsInstance
      >>> 5 == IsInstance(int)
      True
      >>> 'foobar' == IsInstance(int)
      False
      >>> 'foobar' == IsInstance(int, str)
      True
    """
    def __init__(self, *types: Type[Any]):
        super().__init__()
        self.types = types

    def _matches(self, other: Any) -> bool:
        return isinstance(other, self.types)

    def _description(self) -> str:
        return ', '.join(map(lambda t: t.__name__, self.types))


OfType = IsInstance


class ConvertsTo(Matcher):
    """A Matcher checking that the compared value can be converted into the passed type.

    :param type_: The type to which conversion shall be attempted.

    Usage:
      >>> from pychoir import ConvertsTo
      >>> '5' == ConvertsTo(int)
      True
      >>> '5.2' == ConvertsTo(int)
      False
    """
    def __init__(self, type_: Type[Any]):
        super().__init__()
        self.type = type_

    def _matches(self, other: Any) -> bool:
        try:
            converted = self.type(other)
        except:  # noqa: E722
            return False
        return isinstance(converted, self.type)

    def _description(self) -> str:
        return self.type.__name__
